@@ -1,6 +1,4 @@
 # -*- coding: utf-8 -*-
-from __future__ import division, print_function
-
 import functools
 import math
 import os
@@ -12,9 +10,10 @@ import time
 import h5py
 import numpy as np
 import meep as mp
-from . import mode_solver, with_hermitian_epsilon, verbosity
+from . import mode_solver, with_hermitian_epsilon
 from meep.geom import init_do_averaging
 from meep.simulation import get_num_args
+from meep.verbosity_mgr import Verbosity
 try:
     basestring
 except NameError:
@@ -22,8 +21,9 @@ except NameError:
 
 U_MIN = 0
 U_PROD = 1
-U_SUM = 2
+U_MEAN = 2
 
+verbosity = Verbosity(mp.cvar, 'meep', 1)
 
 class MPBArray(np.ndarray):
 
@@ -84,9 +84,9 @@ class ModeSolver(object):
                  target_freq=0.0,
                  tolerance=1.0e-7,
                  num_bands=1,
-                 k_points=[],
+                 k_points=None,
                  ensure_periodicity=True,
-                 geometry=[],
+                 geometry=None,
                  geometry_lattice=mp.Lattice(),
                  geometry_center=mp.Vector3(0, 0, 0),
                  default_material=mp.Medium(epsilon=1),
@@ -102,8 +102,8 @@ class ModeSolver(object):
         self.mode_solver = None
         self.resolution = resolution
         self.eigensolver_flags = eigensolver_flags
-        self.k_points = k_points
-        self.geometry = geometry
+        self.k_points = k_points if k_points else []
+        self.geometry = geometry if geometry else []
         self.geometry_lattice = geometry_lattice
         self.geometry_center = mp.Vector3(*geometry_center)
         self.default_material = default_material
@@ -510,7 +510,7 @@ class ModeSolver(object):
         return update_brd(brd, freqs, [])
 
     def output_band_range_data(self, br_data):
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             for tup, band in zip(br_data, range(1, len(br_data) + 1)):
                 fmt = "Band {} range: {} at {} to {} at {}"
                 min_band, max_band = tup
@@ -537,7 +537,7 @@ class ModeSolver(object):
                 else:
                     gap_size = ((200 * (br_rest_min_f - br_cur_max_f)) /
                                 (br_rest_min_f + br_cur_max_f))
-                    if verbosity.mpb >= 1:
+                    if verbosity.mpb > 0:
                         fmt = "Gap from band {} ({}) to band {} ({}), {}%"
                         print(fmt.format(i, br_cur_max_f, i + 1, br_rest_min_f, gap_size))
                     return ogaps(br_rest[0], br_rest[1:], i + 1,
@@ -642,7 +642,7 @@ class ModeSolver(object):
             self.freqs[curfield_band - 1]
         )
         fname = self._create_fname(fname, fname_prefix, True)
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("Outputting complex scalar field to {}...".format(fname))
 
         with h5py.File(fname, 'w') as f:
@@ -676,7 +676,7 @@ class ModeSolver(object):
         )
 
         fname = self._create_fname(fname, fname_prefix, True)
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("Outputting fields to {}...".format(fname))
 
         with h5py.File(fname, 'w') as f:
@@ -721,7 +721,7 @@ class ModeSolver(object):
 
         parity_suffix = False if curfield_type in 'mn' else True
         fname = self._create_fname(fname, fname_prefix, parity_suffix)
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("Outputting {}...".format(fname))
 
         with h5py.File(fname, 'w') as f:
@@ -810,7 +810,7 @@ class ModeSolver(object):
 
     def display_kpoint_data(self, name, data):
         k_index = self.mode_solver.get_kpoint_index()
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("{}{}:, {}".format(self.parity, name, k_index), end='')
 
             for d in data:
@@ -826,7 +826,7 @@ class ModeSolver(object):
         min_iters = min(self.eigensolver_iters)
         max_iters = max(self.eigensolver_iters)
         mean_iters = np.mean(self.eigensolver_iters)
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             fmt = "eigensolver iterations for {} kpoints: {}-{}, mean = {}"
             print(fmt.format(num_runs, min_iters, max_iters, mean_iters), end='')
 
@@ -834,15 +834,15 @@ class ModeSolver(object):
         idx1 = num_runs // 2
         idx2 = ((num_runs + 1) // 2) - 1
         median_iters = 0.5 * (sorted_iters[idx1] + sorted_iters[idx2])
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print(", median = {}".format(median_iters))
 
         mean_flops = self.eigensolver_flops / (num_runs * mean_iters)
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("mean flops per iteration = {}".format(mean_flops))
 
         mean_time = self.total_run_time / (mean_iters * num_runs)
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("mean time per iteration = {} s".format(mean_time))
 
     def _get_grid_size(self):
@@ -884,6 +884,7 @@ class ModeSolver(object):
 
     def solve_kpoint(self, k):
         self.mode_solver.solve_kpoint(k)
+        self.freqs = self.get_freqs()
 
     def run_parity(self, p, reset_fields, *band_functions):
         if self.random_fields and self.randomize_fields not in band_functions:
@@ -896,7 +897,7 @@ class ModeSolver(object):
 
         init_time = time.time()
 
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("Initializing eigensolver data")
             print("Computing {} bands with {} tolerance".format(self.num_bands, self.tolerance))
 
@@ -905,7 +906,7 @@ class ModeSolver(object):
         if isinstance(reset_fields, basestring):
             self.load_eigenvectors(reset_fields)
 
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("{} k-points".format(len(self.k_points)))
 
             for kp in self.k_points:
@@ -922,11 +923,10 @@ class ModeSolver(object):
             for i, k in enumerate(k_split[1]):
                 self.current_k = k
                 solve_kpoint_time = time.time()
-                self.mode_solver.solve_kpoint(k)
+                self.solve_kpoint(k)
                 self.iterations = self.mode_solver.get_iterations()
-                if verbosity.mpb >= 1:
+                if verbosity.mpb > 0:
                     print("elapsed time for k point: {}".format(time.time() - solve_kpoint_time))
-                self.freqs = self.get_freqs()
                 self.all_freqs[i, :] = np.array(self.freqs)
                 self.band_range_data = self.update_band_range_data(self.band_range_data,
                                                                    self.freqs, k)
@@ -952,12 +952,12 @@ class ModeSolver(object):
                 self.gap_list = []
 
         end = time.time() - start
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("total elapsed time for run: {}".format(end))
         self.total_run_time += end
         self.eigensolver_flops = self.mode_solver.get_eigensolver_flops()
         self.parity = self.mode_solver.get_parity_string()
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("done")
 
     def run(self, *band_functions):
@@ -1024,7 +1024,7 @@ class ModeSolver(object):
                 # First, look in the cached table
                 tab_val = bktab.get((b, k), None)
                 if tab_val:
-                    if verbosity.mpb >= 1:
+                    if verbosity.mpb > 0:
                         print("find-k {} at {}: {} (cached)".format(b, k, tab_val[0]))
                     return tab_val
                 # Otherwise, compute bands and cache results
@@ -1046,7 +1046,7 @@ class ModeSolver(object):
                         bktab[(_b, k)] = (_f - omega, _v)
 
                     fun = self.freqs[-1] - omega
-                    if verbosity.mpb >= 1:
+                    if verbosity.mpb > 0:
                         print("find-k {} at {}: {}".format(b, k, fun))
                     return (fun, v[-1])
 
@@ -1075,7 +1075,7 @@ class ModeSolver(object):
         self.num_bands = num_bands_save
         self.k_points = kpoints_save
         ks = list(reversed(ks))
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             print("{}kvals:, {}, {}, {}".format(self.parity, omega, band_min, band_max), end='')
             for k in korig:
                 print(", {}".format(k), end='')
@@ -1126,6 +1126,18 @@ class ModeSolver(object):
     def get_dominant_planewave(self, band):
         return self.mode_solver.get_dominant_planewave(band)
 
+    def transformed_overlap(self, W, w):
+        return self.mode_solver.transformed_overlap(W, w)
+
+    def compute_symmetry(self, band, W, w):
+        return self.mode_solver.compute_symmetry(band, W, w)
+   
+    def compute_symmetries(self, W, w):
+        symvals = []
+        for band in range(1, self.num_bands+1):
+            symvals.append(self.mode_solver.compute_symmetry(band, W, w))
+        return symvals
+
 
 # Predefined output functions (functions of the band index), for passing to `run`
 
@@ -1239,7 +1251,7 @@ def output_dpwr_in_objects(output_func, min_energy, objects=[]):
         ms.get_dfield(which_band, False)
         ms.compute_field_energy()
         energy = ms.compute_energy_in_objects(objects)
-        if verbosity.mpb >= 1:
+        if verbosity.mpb > 0:
             fmt = "dpwr:, {}, {}, {} "
             print(fmt.format(which_band, ms.freqs[which_band - 1], energy))
         if energy >= min_energy:
