@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2019 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2022 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@
 // Dump/load raw structure data to/from an HDF5 file.  Only
 // works if the number of processors/chunks is the same.
 
+#include <cassert>
 #include <stdio.h>
 #include <stdlib.h>
 #include <math.h>
@@ -32,7 +33,9 @@ namespace meep {
 
 // Write the parameters required to reconstruct the susceptibility (id, noise_amp (for noisy),
 // omega_0, gamma, no_omega_0_denominator)
-void structure::write_susceptibility_params(h5file *file, const char *dname, int EorH) {
+void structure::write_susceptibility_params(h5file *file,
+                                            bool single_parallel_file,
+                                            const char *dname, int EorH) {
   // Get number of susceptibility params from first chunk, since all chunks will have
   // the same susceptibility list.
   size_t params_ntotal = 0;
@@ -44,8 +47,9 @@ void structure::write_susceptibility_params(h5file *file, const char *dname, int
 
   // Write params
   size_t params_start = 0;
-  file->create_data(dname, 1, &params_ntotal);
-  if (am_master()) {
+  file->create_data(dname, 1, &params_ntotal, false /* append_data */,
+                    sizeof(realnum) == sizeof(float) /* single_precision */);
+  if (am_master() || !single_parallel_file) {
     susceptibility *sus = chunks[0]->chiP[EorH];
     while (sus) {
       sus->dump_params(file, &params_start);
@@ -57,10 +61,10 @@ void structure::write_susceptibility_params(h5file *file, const char *dname, int
 void structure::dump_chunk_layout(const char *filename) {
   // Write grid_volume info for each chunk so we can reconstruct chunk division from split_by_cost
   size_t sz = num_chunks * 3;
-  realnum *origins = new realnum[sz];
+  double *origins = new double[sz];
   size_t *nums = new size_t[sz];
   memset(nums, 0, sizeof(size_t) * sz);
-  memset(origins, 0, sizeof(realnum) * sz);
+  memset(origins, 0, sizeof(double) * sz);
   for (int i = 0; i < num_chunks; ++i) {
     int idx = i * 3;
     LOOP_OVER_DIRECTIONS(gv.dim, d) {
@@ -69,7 +73,7 @@ void structure::dump_chunk_layout(const char *filename) {
     }
   }
   h5file file(filename, h5file::WRITE, true);
-  file.create_data("gv_origins", 1, &sz);
+  file.create_data("gv_origins", 1, &sz, false /* append_data */, false /* single_precision */);
   if (am_master()) {
     size_t gv_origins_start = 0;
     file.write_chunk(1, &gv_origins_start, &sz, origins);
@@ -83,39 +87,63 @@ void structure::dump_chunk_layout(const char *filename) {
   delete[] nums;
 }
 
-void structure::dump(const char *filename) {
-  if (verbosity > 0) master_printf("creating epsilon output file \"%s\"...\n", filename);
+void structure::dump(const char *filename, bool single_parallel_file) {
+  if (verbosity > 0) printf("creating epsilon from file \"%s\" (%d)...\n", filename, single_parallel_file);
 
-  // make/save a num_chunks x NUM_FIELD_COMPONENTS x 5 array counting
-  // the number of entries in the chi1inv array for each chunk.
-  size_t *num_chi1inv_ = new size_t[num_chunks * NUM_FIELD_COMPONENTS * 5];
-  memset(num_chi1inv_, 0, sizeof(size_t) * size_t(num_chunks * NUM_FIELD_COMPONENTS * 5));
+  /*
+   * make/save a num_chunks x NUM_FIELD_COMPONENTS x 5 array counting
+   * the number of entries in the chi1inv array for each chunk.
+   *
+   * When 'single_parallel_file' is true, we are creating a single block of data
+   * for ALL chunks (that are merged using MPI). Otherwise, we are just
+   * making a copy of just the chunks that are ours.
+   */
+  int my_num_chunks = 0;
+  for (int i = 0; i < num_chunks; i++) {
+    my_num_chunks += (single_parallel_file || chunks[i]->is_mine());
+  }
+  size_t num_chi1inv_size = my_num_chunks * NUM_FIELD_COMPONENTS * 5;
+  std::vector<size_t> num_chi1inv_(num_chi1inv_size);
   size_t my_ntot = 0;
-  for (int i = 0; i < num_chunks; i++)
+  for (int i = 0, chunk_i = 0; i < num_chunks; i++) {
     if (chunks[i]->is_mine()) {
       size_t ntot = chunks[i]->gv.ntot();
-      for (int c = 0; c < NUM_FIELD_COMPONENTS; ++c)
-        for (int d = 0; d < 5; ++d)
+      for (int c = 0; c < NUM_FIELD_COMPONENTS; ++c) {
+        for (int d = 0; d < 5; ++d) {
           if (chunks[i]->chi1inv[c][d])
-            my_ntot += (num_chi1inv_[(i * NUM_FIELD_COMPONENTS + c) * 5 + d] = ntot);
+            my_ntot += (num_chi1inv_[(chunk_i * NUM_FIELD_COMPONENTS + c) * 5 + d] = ntot);
+        }
+      }
     }
-  size_t *num_chi1inv = new size_t[num_chunks * NUM_FIELD_COMPONENTS * 5];
-  sum_to_master(num_chi1inv_, num_chi1inv, num_chunks * NUM_FIELD_COMPONENTS * 5);
-  delete[] num_chi1inv_;
+    chunk_i += (chunks[i]->is_mine() || single_parallel_file);
+  }
+
+  std::vector<size_t> num_chi1inv;
+  if (single_parallel_file) {
+    num_chi1inv.resize(num_chi1inv_size);
+    sum_to_master(num_chi1inv_.data(), num_chi1inv.data(), num_chi1inv_size);
+  } else {
+    num_chi1inv = std::move(num_chi1inv_);
+  }
 
   // determine total dataset size and offset of this process's data
-  size_t my_start = partial_sum_to_all(my_ntot) - my_ntot;
-  size_t ntotal = sum_to_all(my_ntot);
+  size_t my_start = 0;
+  size_t ntotal = my_ntot;
+  if (single_parallel_file) {
+    my_start = partial_sum_to_all(my_ntot) - my_ntot;
+    ntotal = sum_to_all(my_ntot);
+  }
 
-  h5file file(filename, h5file::WRITE, true);
-  size_t dims[3] = {(size_t)num_chunks, NUM_FIELD_COMPONENTS, 5};
+  h5file file(filename, h5file::WRITE, single_parallel_file, !single_parallel_file);
+  size_t dims[3] = {(size_t)my_num_chunks, NUM_FIELD_COMPONENTS, 5};
   size_t start[3] = {0, 0, 0};
   file.create_data("num_chi1inv", 3, dims);
-  if (am_master()) file.write_chunk(3, start, dims, num_chi1inv);
-  delete[] num_chi1inv;
+  if (am_master() || !single_parallel_file) {
+    file.write_chunk(3, start, dims, num_chi1inv.data());
+  }
 
   // write the data
-  file.create_data("chi1inv", 1, &ntotal);
+  file.create_data("chi1inv", 1, &ntotal, false /* append_data */, false /* single_precision */);
   for (int i = 0; i < num_chunks; i++)
     if (chunks[i]->is_mine()) {
       size_t ntot = chunks[i]->gv.ntot();
@@ -145,7 +173,7 @@ void structure::dump(const char *filename) {
     // Write the number of susceptibilites
     size_t len = 2;
     file.create_data("num_sus", 1, &len);
-    if (am_master()) {
+    if (am_master() || !single_parallel_file) {
       size_t start = 0;
       size_t ntot = 2;
       file.write_chunk(1, &start, &ntot, num_sus);
@@ -155,13 +183,11 @@ void structure::dump(const char *filename) {
   // Get number of non-null sigma entries for each chiP in each chunk.
   // Assumes each susceptibility in the chiP[E_stuff] list has the
   // same number of non-null sigma elements. Likewise for chiP[H_stuff]
-  size_t *my_num_sigmas[2];
-  size_t *num_sigmas[2];
+  std::vector<size_t> my_num_sigmas[2];
+  std::vector<size_t> num_sigmas[2];
   for (int ft = 0; ft < 2; ++ft) {
-    my_num_sigmas[ft] = new size_t[num_chunks];
-    num_sigmas[ft] = new size_t[num_chunks];
-    memset(my_num_sigmas[ft], 0, sizeof(size_t) * num_chunks);
-    memset(num_sigmas[ft], 0, sizeof(size_t) * num_chunks);
+    my_num_sigmas[ft].resize(num_chunks);
+    num_sigmas[ft].resize(num_chunks);
   }
 
   for (int i = 0; i < num_chunks; ++i) {
@@ -197,7 +223,7 @@ void structure::dump(const char *filename) {
 
   file.prevent_deadlock();
   for (int ft = 0; ft < 2; ++ft) {
-    sum_to_all(my_num_sigmas[ft], num_sigmas[ft], num_chunks);
+    sum_to_all(my_num_sigmas[ft].data(), num_sigmas[ft].data(), num_chunks);
   }
 
   size_t num_E_sigmas = 0;
@@ -208,17 +234,13 @@ void structure::dump(const char *filename) {
   }
 
   // Allocate space for component and direction of non-null sigmas
-  size_t *my_sigma_cd[2] = {NULL, NULL};
-  my_sigma_cd[E_stuff] = new size_t[num_E_sigmas * 2];
-  memset(my_sigma_cd[E_stuff], 0, sizeof(size_t) * num_E_sigmas * 2);
-  my_sigma_cd[H_stuff] = new size_t[num_H_sigmas * 2];
-  memset(my_sigma_cd[H_stuff], 0, sizeof(size_t) * num_H_sigmas * 2);
+  std::vector<size_t> my_sigma_cd[2];
+  my_sigma_cd[E_stuff].resize(num_E_sigmas * 2);
+  my_sigma_cd[H_stuff].resize(num_H_sigmas * 2);
 
-  size_t *sigma_cd[2] = {NULL, NULL};
-  sigma_cd[E_stuff] = new size_t[num_E_sigmas * 2];
-  memset(sigma_cd[E_stuff], 0, sizeof(size_t) * num_E_sigmas * 2);
-  sigma_cd[H_stuff] = new size_t[num_H_sigmas * 2];
-  memset(sigma_cd[H_stuff], 0, sizeof(size_t) * num_H_sigmas * 2);
+  std::vector<size_t> sigma_cd[2];
+  sigma_cd[E_stuff].resize(num_E_sigmas * 2);
+  sigma_cd[H_stuff].resize(num_H_sigmas * 2);
 
   // Find component and direction of non-null sigmas
   {
@@ -244,8 +266,8 @@ void structure::dump(const char *filename) {
         }
       }
     }
-    bw_or_to_all(my_sigma_cd[E_stuff], sigma_cd[E_stuff], num_E_sigmas * 2);
-    bw_or_to_all(my_sigma_cd[H_stuff], sigma_cd[H_stuff], num_H_sigmas * 2);
+    bw_or_to_all(my_sigma_cd[E_stuff].data(), sigma_cd[E_stuff].data(), num_E_sigmas * 2);
+    bw_or_to_all(my_sigma_cd[H_stuff].data(), sigma_cd[H_stuff].data(), num_H_sigmas * 2);
   }
 
   // Write location (component and direction) data of non-null sigmas (sigma[c][d])
@@ -254,9 +276,9 @@ void structure::dump(const char *filename) {
     file.create_data("sigma_cd", 1, &len);
     size_t start = 0;
     for (int ft = 0; ft < 2; ++ft) {
-      if (am_master()) {
-        size_t count = (ft == 0 ? num_E_sigmas : num_H_sigmas) * 2;
-        file.write_chunk(1, &start, &count, sigma_cd[ft]);
+      if (am_master() || !single_parallel_file) {
+        size_t count = sigma_cd[ft].size();
+        file.write_chunk(1, &start, &count, sigma_cd[ft].data());
         start += count;
       }
     }
@@ -290,15 +312,8 @@ void structure::dump(const char *filename) {
     }
   }
 
-  write_susceptibility_params(&file, "E_params", E_stuff);
-  write_susceptibility_params(&file, "H_params", H_stuff);
-
-  for (int ft = 0; ft < 2; ++ft) {
-    delete[] sigma_cd[ft];
-    delete[] my_sigma_cd[ft];
-    delete[] num_sigmas[ft];
-    delete[] my_num_sigmas[ft];
-  }
+  write_susceptibility_params(&file, single_parallel_file, "E_params", E_stuff);
+  write_susceptibility_params(&file, single_parallel_file, "H_params", H_stuff);
 }
 
 // Reconstruct the chiP lists of susceptibilities from the params hdf5 data
@@ -312,7 +327,6 @@ susceptibility *make_sus_list_from_params(h5file *file, int rank, size_t dims[3]
     realnum num_params;
     file->read_chunk(rank, &start, num_params_dims, &num_params);
     start += num_params_dims[0];
-
     if (num_params == 4) {
       // This is a lorentzian_susceptibility and the next 4 values in the dataset
       // are id, omega_0, gamma, and no_omega_0_denominator.
@@ -394,7 +408,7 @@ susceptibility *make_sus_list_from_params(h5file *file, int rank, size_t dims[3]
       if (sus->next) sus = sus->next;
     }
     else {
-      abort("Invalid number of susceptibility parameters in structure::load");
+      meep::abort("Invalid number of susceptibility parameters in structure::load");
     }
   }
   return res;
@@ -405,7 +419,7 @@ void structure::set_chiP_from_file(h5file *file, const char *dataset, field_type
   size_t dims[3] = {0, 0, 0};
 
   file->read_size(dataset, &rank, dims, 1);
-  if (rank != 1) abort("inconsistent data size in structure::load");
+  if (rank != 1) meep::abort("inconsistent data size in structure::set_chiP_from_file");
 
   if (dims[0] != 0) {
     for (int i = 0; i < num_chunks; ++i) {
@@ -414,19 +428,80 @@ void structure::set_chiP_from_file(h5file *file, const char *dataset, field_type
   }
 }
 
+binary_partition::binary_partition(int _proc_id) : proc_id(_proc_id), plane{NO_DIRECTION, 0.0} {}
+
+binary_partition::binary_partition(const split_plane &_split_plane,
+                                   std::unique_ptr<binary_partition> &&left_tree,
+                                   std::unique_ptr<binary_partition> &&right_tree)
+    : proc_id(-1), plane(_split_plane), left(std::move(left_tree)), right(std::move(right_tree)) {
+  if (!left || !right) { meep::abort("Binary partition tree is required to be full"); }
+}
+
+binary_partition::binary_partition(const binary_partition& other) : proc_id(other.proc_id), plane(other.plane) {
+  if (!other.is_leaf()) {
+    left.reset(new binary_partition(*other.left));
+    right.reset(new binary_partition(*other.right));
+  }
+}
+
+bool binary_partition::is_leaf() const { return !left && !right; }
+
+int binary_partition::get_proc_id() const {
+  assert(is_leaf());
+  return proc_id;
+}
+
+const split_plane &binary_partition::get_plane() const {
+  assert(!is_leaf());
+  return plane;
+}
+
+const binary_partition *binary_partition::left_tree() const {
+  assert(!is_leaf());
+  return left.get();
+}
+
+const binary_partition *binary_partition::right_tree() const {
+  assert(!is_leaf());
+  return right.get();
+}
+
+void split_by_binarytree(grid_volume gvol, std::vector<grid_volume> &result_gvs,
+                         std::vector<int> &result_ids, const binary_partition *bp) {
+  // reached a leaf
+  if (bp->is_leaf()) {
+    result_gvs.push_back(gvol);
+    result_ids.push_back(bp->get_proc_id());
+    return;
+  }
+
+  const auto &plane = bp->get_plane();
+  int split_point = static_cast<int>(
+      (plane.pos - gvol.surroundings().in_direction_min(plane.dir)) /
+          gvol.surroundings().in_direction(plane.dir) * gvol.num_direction(plane.dir) +
+      0.5);
+  // traverse left branch
+  grid_volume left_gvol = gvol.split_at_fraction(false, split_point, plane.dir);
+  split_by_binarytree(left_gvol, result_gvs, result_ids, bp->left_tree());
+
+  // traverse right branch
+  grid_volume right_gvol = gvol.split_at_fraction(true, split_point, plane.dir);
+  split_by_binarytree(right_gvol, result_gvs, result_ids, bp->right_tree());
+}
+
 void structure::load_chunk_layout(const char *filename, boundary_region &br) {
   // Load chunk grid_volumes from a file
   h5file file(filename, h5file::READONLY, true);
   size_t sz = num_chunks * 3;
-  realnum *origins = new realnum[sz];
-  memset(origins, 0, sizeof(realnum) * sz);
+  double *origins = new double[sz];
+  memset(origins, 0, sizeof(double) * sz);
   size_t *nums = new size_t[sz];
   memset(nums, 0, sizeof(size_t) * sz);
 
   int origins_rank;
   size_t origins_dims;
   file.read_size("gv_origins", &origins_rank, &origins_dims, 1);
-  if (origins_rank != 1 || origins_dims != sz) { abort("chunk mismatch in structure::load"); }
+  if (origins_rank != 1 || origins_dims != sz) { meep::abort("chunk mismatch in structure::load"); }
   if (am_master()) {
     size_t gv_origins_start = 0;
     file.read_chunk(1, &gv_origins_start, &origins_dims, origins);
@@ -437,7 +512,7 @@ void structure::load_chunk_layout(const char *filename, boundary_region &br) {
   int nums_rank;
   size_t nums_dims;
   file.read_size("gv_nums", &nums_rank, &nums_dims, 1);
-  if (nums_rank != 1 || nums_dims != sz) { abort("chunk mismatch in structure::load"); }
+  if (nums_rank != 1 || nums_dims != sz) { meep::abort("chunk mismatch in structure::load"); }
   if (am_master()) {
     size_t gv_nums_start = 0;
     file.read_chunk(1, &gv_nums_start, &nums_dims, nums);
@@ -446,6 +521,7 @@ void structure::load_chunk_layout(const char *filename, boundary_region &br) {
   broadcast(0, nums, sz);
 
   std::vector<grid_volume> gvs;
+  std::vector<int> ids;
   // Populate a vector with the new grid_volumes
   for (int i = 0; i < num_chunks; ++i) {
     int idx = i * 3;
@@ -457,73 +533,101 @@ void structure::load_chunk_layout(const char *filename, boundary_region &br) {
     }
     new_gv.set_origin(new_origin);
     gvs.push_back(new_gv);
+    ids.push_back(i * count_processors() / num_chunks);
   }
 
-  load_chunk_layout(gvs, br);
+  load_chunk_layout(gvs, ids, br);
 
   delete[] origins;
   delete[] nums;
 }
 
-void structure::load_chunk_layout(const std::vector<grid_volume> &gvs, boundary_region &br) {
+void structure::load_chunk_layout(const std::vector<grid_volume> &gvs,
+                                  const std::vector<int> &ids,
+                                  boundary_region &br) {
+  if (gvs.size() != size_t(num_chunks)) meep::abort("load_chunk_layout: wrong number of chunks.");
   // Recreate the chunks with the new grid_volumes
   for (int i = 0; i < num_chunks; ++i) {
     if (chunks[i]->refcount-- <= 1) delete chunks[i];
-    int proc = i * count_processors() / num_chunks;
-    chunks[i] = new structure_chunk(gvs[i], v, Courant, proc);
+    chunks[i] = new structure_chunk(gvs[i], v, Courant, ids[i] % count_processors());
     br.apply(this, chunks[i]);
   }
   check_chunks();
 }
 
-void structure::load(const char *filename) {
-  h5file file(filename, h5file::READONLY, true);
+void structure::load(const char *filename, bool single_parallel_file) {
+  h5file file(filename, h5file::READONLY, single_parallel_file, !single_parallel_file);
 
-  if (verbosity > 0) master_printf("reading epsilon from file \"%s\"...\n", filename);
+  if (verbosity > 0) printf("reading epsilon from file \"%s\" (%d)...\n", filename, single_parallel_file);
+
+  /*
+   * make/save a num_chunks x NUM_FIELD_COMPONENTS x 5 array counting
+   * the number of entries in the chi1inv array for each chunk.
+   *
+   * When 'single_parallel_file' is true, we are creating a single block of data
+   * for ALL chunks (that are merged using MPI). Otherwise, we are just
+   * making a copy of just the chunks that are ours.
+   */
+  int my_num_chunks = 0;
+  for (int i = 0; i < num_chunks; i++) {
+    my_num_chunks += (single_parallel_file || chunks[i]->is_mine());
+  }
+  size_t num_chi1inv_size = my_num_chunks * NUM_FIELD_COMPONENTS * 5;
+  std::vector<size_t> num_chi1inv(num_chi1inv_size);
 
-  // make/save a num_chunks x NUM_FIELD_COMPONENTS x 5 array counting
-  // the number of entries in the chi1inv array for each chunk.
-  size_t *num_chi1inv = new size_t[num_chunks * NUM_FIELD_COMPONENTS * 5];
   int rank;
-  size_t dims[3], _dims[3] = {(size_t)num_chunks, NUM_FIELD_COMPONENTS, 5};
+  size_t dims[3], _dims[3] = {(size_t)my_num_chunks, NUM_FIELD_COMPONENTS, 5};
   size_t start[3] = {0, 0, 0};
   file.read_size("num_chi1inv", &rank, dims, 3);
   if (rank != 3 || _dims[0] != dims[0] || _dims[1] != dims[1] || _dims[2] != dims[2])
-    abort("chunk mismatch in structure::load");
-  if (am_master()) file.read_chunk(3, start, dims, num_chi1inv);
+    meep::abort("chunk mismatch in structure::load");
+  if (am_master() || !single_parallel_file) file.read_chunk(3, start, dims, num_chi1inv.data());
 
-  file.prevent_deadlock();
-  broadcast(0, num_chi1inv, dims[0] * dims[1] * dims[2]);
+  if (single_parallel_file) {
+    file.prevent_deadlock();
+    broadcast(0, num_chi1inv.data(), dims[0] * dims[1] * dims[2]);
+  }
 
   changing_chunks();
 
   // allocate data as needed and check sizes
   size_t my_ntot = 0;
-  for (int i = 0; i < num_chunks; i++)
+  for (int i = 0, chunk_i = 0; i < num_chunks; i++) {
     if (chunks[i]->is_mine()) {
       size_t ntot = chunks[i]->gv.ntot();
       for (int c = 0; c < NUM_FIELD_COMPONENTS; ++c)
         for (int d = 0; d < 5; ++d) {
-          size_t n = num_chi1inv[(i * NUM_FIELD_COMPONENTS + c) * 5 + d];
+          size_t n = num_chi1inv[(chunk_i * NUM_FIELD_COMPONENTS + c) * 5 + d];
           if (n == 0) {
             delete[] chunks[i]->chi1inv[c][d];
             chunks[i]->chi1inv[c][d] = NULL;
           }
           else {
-            if (n != ntot) abort("grid size mismatch %zd vs %zd in structure::load", n, ntot);
-            chunks[i]->chi1inv[c][d] = new realnum[ntot];
+            if (n != ntot) meep::abort("grid size mismatch %zd vs %zd in structure::load", n, ntot);
+            if (!chunks[i]->chi1inv[c][d]) chunks[i]->chi1inv[c][d] = new realnum[ntot];
             my_ntot += ntot;
           }
         }
     }
+    chunk_i += (chunks[i]->is_mine() || single_parallel_file);
+  }
 
   // determine total dataset size and offset of this process's data
-  size_t my_start = partial_sum_to_all(my_ntot) - my_ntot;
-  size_t ntotal = sum_to_all(my_ntot);
+  size_t my_start = 0;
+  size_t ntotal = my_ntot;
+  if (single_parallel_file) {
+    my_start = partial_sum_to_all(my_ntot) - my_ntot;
+    ntotal = sum_to_all(my_ntot);
+  }
 
   // read the data
   file.read_size("chi1inv", &rank, dims, 1);
-  if (rank != 1 || dims[0] != ntotal) abort("inconsistent data size in structure::load");
+  if (rank != 1 || dims[0] != ntotal) {
+    meep::abort(
+        "inconsistent data size for chi1inv in structure::load (rank, dims[0]): "
+        "(%d, %zu) != (1, %zu)",
+        rank, dims[0], ntotal);
+  }
   for (int i = 0; i < num_chunks; i++)
     if (chunks[i]->is_mine()) {
       size_t ntot = chunks[i]->gv.ntot();
@@ -545,44 +649,44 @@ void structure::load(const char *filename) {
     size_t dims[] = {0, 0, 0};
     file.read_size("num_sus", &rank, dims, 1);
     if (dims[0] > 0) {
-      if (am_master()) {
+      if (am_master() || !single_parallel_file) {
         size_t start = 0;
         size_t count = 2;
         file.read_chunk(rank, &start, &count, num_sus);
       }
     }
-    file.prevent_deadlock();
-    broadcast(0, num_sus, 2);
+    if (single_parallel_file) {
+      file.prevent_deadlock();
+      broadcast(0, num_sus, 2);
+    }
   }
 
   // Allocate and read non-null sigma entry data
-  size_t *num_sigmas[2] = {NULL, NULL};
-  num_sigmas[E_stuff] = new size_t[num_chunks];
-  memset(num_sigmas[E_stuff], 0, sizeof(size_t) * num_chunks);
-  num_sigmas[H_stuff] = new size_t[num_chunks];
-  memset(num_sigmas[H_stuff], 0, sizeof(size_t) * num_chunks);
+  std::vector<size_t> num_sigmas[2];
+  num_sigmas[E_stuff].resize(num_chunks);
+  num_sigmas[H_stuff].resize(num_chunks);
 
   // Read num_sigmas data
   {
     int rank = 0;
     size_t dims[] = {0, 0, 0};
     file.read_size("num_sigmas", &rank, dims, 1);
-    if (dims[0] != (size_t)num_chunks * 2) { abort("inconsistent data size in structure::load"); }
-    if (am_master()) {
+    if (dims[0] != (size_t)num_chunks * 2) { meep::abort("inconsistent data size for num_sigmas in structure::load"); }
+    if (am_master() || !single_parallel_file) {
       size_t start = 0;
       size_t count = num_chunks;
-      file.read_chunk(rank, &start, &count, num_sigmas[E_stuff]);
+      file.read_chunk(rank, &start, &count, num_sigmas[E_stuff].data());
       start += count;
-      file.read_chunk(rank, &start, &count, num_sigmas[H_stuff]);
+      file.read_chunk(rank, &start, &count, num_sigmas[H_stuff].data());
+    }
+    if (single_parallel_file) {
+      file.prevent_deadlock();
+      broadcast(0, num_sigmas[E_stuff].data(), num_chunks);
+      broadcast(0, num_sigmas[H_stuff].data(), num_chunks);
     }
-    file.prevent_deadlock();
-    broadcast(0, num_sigmas[E_stuff], num_chunks);
-    broadcast(0, num_sigmas[H_stuff], num_chunks);
   }
 
   // Allocate space for component and direction data of the non-null susceptibilities
-  size_t *sigma_cd[2] = {NULL, NULL};
-
   size_t nsig[2] = {0, 0};
   for (int ft = 0; ft < 2; ++ft) {
     for (int i = 0; i < num_chunks; ++i) {
@@ -590,10 +694,9 @@ void structure::load(const char *filename) {
     }
   }
 
-  sigma_cd[E_stuff] = new size_t[nsig[E_stuff] * 2];
-  memset(sigma_cd[E_stuff], 0, sizeof(size_t) * nsig[E_stuff] * 2);
-  sigma_cd[H_stuff] = new size_t[nsig[H_stuff] * 2];
-  memset(sigma_cd[H_stuff], 0, sizeof(size_t) * nsig[H_stuff] * 2);
+  std::vector<size_t> sigma_cd[2];
+  sigma_cd[E_stuff].resize(nsig[E_stuff] * 2);
+  sigma_cd[H_stuff].resize(nsig[H_stuff] * 2);
 
   // Read the component/direction data
   {
@@ -601,20 +704,22 @@ void structure::load(const char *filename) {
     size_t dims[] = {0, 0, 0};
     file.read_size("sigma_cd", &rank, dims, 1);
     if (dims[0] != 2 * (nsig[E_stuff] + nsig[H_stuff])) {
-      abort("inconsistent data size in structure::load");
+      meep::abort("inconsistent data size for sigma_cd in structure::load");
     }
 
-    if (am_master()) {
+    if (am_master() || !single_parallel_file) {
       size_t start = 0;
       for (int ft = 0; ft < 2; ++ft) {
-        size_t count = nsig[ft] * 2;
-        file.read_chunk(rank, &start, &count, sigma_cd[ft]);
+        size_t count = sigma_cd[ft].size();
+        file.read_chunk(rank, &start, &count, sigma_cd[ft].data());
         start += count;
       }
     }
-    file.prevent_deadlock();
-    broadcast(0, sigma_cd[E_stuff], nsig[E_stuff] * 2);
-    broadcast(0, sigma_cd[H_stuff], nsig[H_stuff] * 2);
+    if (single_parallel_file) {
+      file.prevent_deadlock();
+      broadcast(0, sigma_cd[E_stuff].data(), nsig[E_stuff] * 2);
+      broadcast(0, sigma_cd[H_stuff].data(), nsig[H_stuff] * 2);
+    }
   }
 
   for (int ft = 0; ft < 2; ++ft) {
@@ -664,10 +769,5 @@ void structure::load(const char *filename) {
       }
     }
   }
-
-  for (int ft = 0; ft < 2; ++ft) {
-    delete[] num_sigmas[ft];
-    delete[] sigma_cd[ft];
-  }
 }
 } // namespace meep
