@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2021 Massachusetts Institute of Technology.
+/* Copyright (C) 2005-2022 Massachusetts Institute of Technology.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -224,10 +224,10 @@ void doit(bool use_hdf5) {
     if (errT > maxerrT) maxerrT = errT;
     if (errR > maxerrR) maxerrR = errR;
     if (errT * sqr(freq_min / (freq_min + i * dfreq)) > 0.01)
-      abort("large error %g at freq = %g: T = %g instead of %g\n", errT, freq_min + i * dfreq, T[i],
+      meep::abort("large error %g at freq = %g: T = %g instead of %g\n", errT, freq_min + i * dfreq, T[i],
             T0[i]);
     if (errR * sqr(freq_min / (freq_min + i * dfreq)) > 0.01)
-      abort("large error %g at freq = %g: R = %g instead of %g\n", errR, freq_min + i * dfreq, R[i],
+      meep::abort("large error %g at freq = %g: R = %g instead of %g\n", errR, freq_min + i * dfreq, R[i],
             R0[i]);
   }
   master_printf("Done (max. err in T = %e, in R = %e)\n", maxerrT, maxerrR);
